#ifndef LTRACE_ELF_H
#define LTRACE_ELF_H

#include <gelf.h>
#include <stdlib.h>

#include "ltrace.h"

struct ltelf {
	int fd;
	Elf *elf;
	GElf_Ehdr ehdr;
	Elf_Data *dynsym;
	size_t dynsym_count;
	const char *dynstr;
	const char *soname;
	const char *interp;
	GElf_Addr plt_addr;
	size_t plt_size;
	Elf_Data *relplt;
	size_t relplt_count;
	Elf_Data *symtab;
	const char *strtab;
	size_t symtab_count;
	Elf_Data *opd;
	GElf_Addr *opd_addr;
	size_t opd_size;
	Elf32_Word *hash;
	int hash_type;
	int lte_flags;
	GElf_Addr dyn_addr;
	size_t dyn_sz;
	GElf_Addr plt_stub_vma;
	GElf_Addr bias;
};

#define ELF_MAX_SEGMENTS  50
#define LTE_HASH_MALLOCED 1
#define LTE_PLT_EXECUTABLE 2

#define PLTS_ARE_EXECUTABLE(lte) ((lte->lte_flags & LTE_PLT_EXECUTABLE) != 0)

extern int library_num;
extern char *library[MAX_LIBRARY];

extern void open_elf(struct ltelf *lte, const char *filename);
extern struct library_symbol *read_elf(struct process *, GElf_Addr *entryp);

extern GElf_Addr arch_plt_sym_val(struct ltelf *, size_t, GElf_Rela *);
extern GElf_Addr opd2addr(struct ltelf *ltc, void *addr);

extern struct ltelf main_lte;
extern int in_load_libraries(const char *name, struct ltelf *lte, size_t count, GElf_Sym *sym);
extern void add_library_symbol(GElf_Addr addr, const char *name,
		struct library_symbol **library_symbolspp,
		enum toplt type_of_plt, int is_weak);

#endif
