#if HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sys/param.h>
#include <signal.h>
#include <sys/wait.h>
#include <sys/ioctl.h>

#include "ltrace.h"
#include "output.h"
#include "read_config_file.h"
#include "options.h"
#include "debug.h"
#include "elf.h"

#ifndef SYSCONFDIR
#define SYSCONFDIR "/etc"
#endif

char *command = NULL;

int exiting = 0;		/* =1 if a SIGINT or SIGTERM has been received */

static enum pcb_status
stop_non_p_processes (struct process * proc, void * data)
{
	int stop = 1;

	struct opt_p_t *it;
	for (it = opt_p; it != NULL; it = it->next) {
		struct process * p_proc = pid2proc(it->pid);
		if (p_proc == NULL) {
			printf("stop_non_p_processes: %d terminated?\n", it->pid);
			continue;
		}
		if (p_proc == proc || p_proc->leader == proc->leader) {
			stop = 0;
			break;
		}
	}

	if (stop) {
		debug(2, "Sending SIGSTOP to process %u", proc->pid);
		kill(proc->pid, SIGSTOP);
	}

	return pcb_cont;
}

static void
signal_alarm(int sig) {
	signal(SIGALRM, SIG_DFL);
	each_process(NULL, &stop_non_p_processes, NULL);
}

static void signal_exit(int sig)
{
	if (exiting != 0)
		return;

	exiting = 1;
	if (should_call_ltrace_exiting())
		check_ltrace_exiting();

	debug(1, "Received interrupt signal; exiting...");
	signal(SIGINT, SIG_IGN);
	signal(SIGTERM, SIG_IGN);
	signal(SIGALRM, signal_alarm);
}

static void normal_exit(void)
{
	output_line(0, 0);
	if (opt_c) {
		show_summary();
	}
}

static void guess_cols(void)
{
	struct winsize ws;
	char *c;

	opt_a = DEFAULT_ACOLUMN;
	c = getenv("COLUMNS");
	if (c && *c) {
		char *endptr;
		int cols;
		cols = strtol(c, &endptr, 0);
		if (cols > 0 && !*endptr) {
			opt_a = cols * 5 / 8;
		}
	} else if (ioctl(1, TIOCGWINSZ, &ws) != -1 && ws.ws_col > 0) {
		opt_a = ws.ws_col * 5 / 8;
	}
}

int main(int argc, char **argv)
{
	struct opt_p_t *opt_p_tmp;
	char *home;

	atexit(normal_exit);
	signal(SIGINT, signal_exit);	/* Detach processes when interrupted */
	signal(SIGTERM, signal_exit);	/*  ... or killed */

	guess_cols();
	argv = process_options(argc, argv);
	read_config_file(SYSCONFDIR "/ltrace.conf");
	home = getenv("HOME");
	if (home) {
		char path[PATH_MAX];
		if (strlen(home) > PATH_MAX - 15) {
			fprintf(stderr, "Error: $HOME too long\n");
			exit(1);
		}
		strcpy(path, getenv("HOME"));
		strcat(path, "/.ltrace.conf");
		read_config_file(path);
	}
	if (opt_e) {
		struct opt_e_t *tmp = opt_e;
		while (tmp) {
			debug(1, "Option -e: %s\n", tmp->name);
			tmp = tmp->next;
		}
	}
	if (command) {
		/* Check that the binary ABI is supported before
		 * calling execute_program.  */
		struct ltelf lte = {};
		open_elf(&lte, command);

		open_program(command, execute_program(command, argv), 0);
	}
	opt_p_tmp = opt_p;
	while (opt_p_tmp) {
		open_pid(opt_p_tmp->pid, 1);
		opt_p_tmp = opt_p_tmp->next;
	}
	while (1) {
		process_event(wait_for_something());
	}
}
